#include "circuit.h"


int Gate::cal_score() {
    score_value_unsatisfied_cost = cal_value_unsatisfied_cost();
    score_fault_propagated_weight[0] = cal_score_fault_propagated_weight(0);
    score_fault_propagated_weight[1] = cal_score_fault_propagated_weight(1);

    score_fault_propagated_unsatisfied_cost[0] = cal_score_fault_propagated_unsatisfied_cost(0);
    score_fault_propagated_unsatisfied_cost[1] = cal_score_fault_propagated_unsatisfied_cost(1);

    score_fault_detected_weight[0] = cal_score_fault_detected_weight(0);
    score_fault_detected_weight[1] = cal_score_fault_detected_weight(1);

    score_fault_detected_unsatisfied_cost[0] = cal_score_fault_detected_unsatisfied_cost(0);
    score_fault_detected_unsatisfied_cost[1] = cal_score_fault_detected_unsatisfied_cost(1);

    return - score_value_unsatisfied_cost 
    + score_fault_propagated_weight[0] + score_fault_propagated_weight[1]
    - score_fault_propagated_unsatisfied_cost[0] - score_fault_propagated_unsatisfied_cost[1]
    + score_fault_detected_weight[0] + score_fault_detected_weight[1]
    - score_fault_detected_unsatisfied_cost[0] - score_fault_detected_unsatisfied_cost[1];
}

int Gate::cal_value_unsatisfied_cost() {
    int res = 0;
    value ^= 1;
    for(Gate* g : fanouts) {
        if(g->value_satisfied && g->cal_value() != g->value) {
            res += g->value_unsatisfied_cost;
        }
        if(!g->value_satisfied && g->cal_value() == g->value) {
            res -= g->value_unsatisfied_cost;
        }
    }
    if(this->value_satisfied) {
        res += this->value_unsatisfied_cost;
    } else {
        res -= this->value_unsatisfied_cost;
    }
    value ^= 1;
    return res;
}

int Gate::cal_score_fault_propagated_weight(int sa) {
    int res = 0;
    value ^= 1;

    int origin_fault_propagated_len = this->fault_propagated_len[sa];
    this->fault_propagated_len[sa] = this->cal_propagate_len(sa);

    res += (this->fault_propagated_len[sa] - origin_fault_propagated_len ) * this->fault_propagated_weight[sa];
    for(Gate* g : fanins) {
        res += ( g->cal_propagate_len(sa) - g->fault_propagated_len[sa] ) * g->fault_propagated_weight[sa];
    }

    this->fault_propagated_len[sa] = origin_fault_propagated_len;

    value ^= 1;
    return res;
}

int Gate::cal_score_fault_propagated_unsatisfied_cost(int sa) {
    int res = 0;
    value ^= 1;
    int origin_fault_propagated_len = this->fault_propagated_len[sa];
    this->fault_propagated_len[sa] = this->cal_propagate_len(sa);

    if(!this->fault_propagated_satisfied[sa]) {
        res -= this->fault_propagated_unsatisfied_cost[res];
    }

    for(Gate* g : fanins) {
        if(g->fault_propagated_satisfied[sa] && g->cal_propagate_len(sa) != g->fault_propagated_len[sa]) {
            res += g->fault_propagated_unsatisfied_cost[sa];
        }
        if(!g->fault_propagated_satisfied[sa] && g->cal_propagate_len(sa) == g->fault_propagated_len[sa]) {
            res -= g->fault_propagated_unsatisfied_cost[sa];
        }
    }

    this->fault_propagated_len[sa] = origin_fault_propagated_len;
    value ^= 1;
    return res;
}

int Gate::cal_score_fault_detected_weight(int sa) {
    int res = 0;
    value ^= 1;

    int origin_fault_detected = this->fault_detected[sa];
    this->fault_detected[sa] = this->cal_fault_detected(sa);
    res += ( this->fault_detected[sa] - origin_fault_detected ) * this->fault_detected_weight[sa];

    for(Gate* g : fanins) {
        res += ( g->cal_fault_detected(sa) - g->fault_detected[sa] ) * this->fault_detected_weight[sa];
    }

    this->fault_detected[sa] = origin_fault_detected;

    value ^= 1;
    return res;
}

int Gate::cal_score_fault_detected_unsatisfied_cost(int sa) {
    int res = 0;
    value ^= 1;

    int origin_fault_detected = this->fault_detected[sa];
    this->fault_detected[sa] = this->cal_fault_detected(sa);

    if(!this->fault_detected_satisfied[sa]) {
        res -= this->fault_detected_unsatisfied_cost[sa];
    }

    for(Gate* g : fanins) {
        if(g->fault_detected_satisfied[sa] && g->cal_fault_detected(sa) != g->fault_detected[sa]) {
            res += g->fault_detected_unsatisfied_cost[sa];
        }
        if(!g->fault_detected_satisfied[sa] && g->cal_fault_detected(sa) == g->fault_detected[sa]) {
            res -= g->fault_detected_unsatisfied_cost[sa];
        }
    }

    this->fault_detected[sa] = origin_fault_detected;

    value ^= 1;
    return res;
}